#!/usr/bin/env python3

import sys
import os
import subprocess
import multiprocessing
import signal
import time
import random

n = None
k = None
w = None
s = None
H = []

reading = None

for line in sys.stdin:
  line = line.strip()

  if line.startswith('#'):
    reading = None
    if line.startswith('# n'):
      reading = 'n'
    elif line.startswith('# k'):
      reading = 'k'
    elif line.startswith('# w'):
      reading = 'w'
    elif line.startswith('# H'):
      reading = 'H'
    elif line.startswith('# s'):
      reading = 's'
    continue

  if reading == 'n':
    n = int(line)
    reading = None
  elif reading == 'k':
    k = int(line)
    reading = None
  elif reading == 'w':
    w = int(line)
    reading = None
  elif reading == 'H':
    assert len(line) == n-k
    H += [line]
  elif reading == 's':
    assert len(line) == n-k
    s = line
    reading = None

assert n != None
assert k != None
assert w != None
assert s != None
assert len(H) == k

if w < 4:
  raise ValueError('sorry, this script supports only w >= 4')

# goal is to find w entries of I|H with sum s

def spaced(s):
  return ' '.join(s)

G = '['
for i in range(k):
  G += '['+spaced(H[i]+'0'*i+'1'+'0'*(k-1-i))+']\n'
G += ']\n'
G += '['+spaced(s+'0'*k)+']\n'

# will feed this (k+1)xn matrix G to lowweight program
# which will search for weight-w sum of rows of G
# in particular: if s is sum of H[i] for w values of i
# then sum of corresponding w rows of G and last row has weight w

# compared to telling search that s must be included,
# this is marginally less efficient, as noted in (e.g.) blp 2008

# reasonable parameter selection:
x = (k+1)//2
l = min(22,n-k)
m = 1
c = 8
# can do better in general by tuning for CPU and (n,k,w)

x = int(os.getenv('ATTACK_X',default=str(x)))
l = int(os.getenv('ATTACK_L',default=str(l)))
m = int(os.getenv('ATTACK_M',default=str(m)))
c = int(os.getenv('ATTACK_C',default=str(c)))

source = 'lowweight.c'
binary = 'binaries/lowweight,n=%d,k=%d,x=%d,l=%d,m=%d,c=%d' % (n,k+1,x,l,m,c)

if not os.path.exists(binary) or os.path.getmtime(binary) < os.path.getmtime(source):
  sys.stderr.write('compiling %s...\n' % binary)
  sys.stderr.flush()

  os.makedirs('binaries',exist_ok=True)
  p = subprocess.run([
    'clang',
    '-g','-O3','-march=native','-mtune=native','-funroll-loops',
    '-Wno-ignored-attributes',
    '-Dn=%d' % n,
    '-Dk=%d' % (k+1),
    '-Dx=%d' % x,
    '-Dl=%d' % l,
    '-Dm=%d' % m,
    '-Dc=%d' % c,
    '-o',binary+'.tmp',
    source
  ],check=True,stdout=subprocess.PIPE,stderr=subprocess.STDOUT,universal_newlines=True)
  sys.stderr.write(p.stdout)

  os.rename(binary+'.tmp',binary)

try:
  threads = len(os.sched_getaffinity(0))
except:
  threads = multiprocessing.cpu_count()
threads = os.getenv('THREADS',default=threads)
threads = int(threads)
if threads < 1: threads = 1

maxiterations = os.getenv('ATTACK_MAX',default=-1)

def lowweight(i):
  r = random.randrange(2**31)
  p = subprocess.Popen(
    ['./'+binary,str(w),str(i),str(r),str(maxiterations)],
    stdin=subprocess.PIPE,
    stdout=subprocess.PIPE,
    universal_newlines=True
  )
  def sigterm(*args):
    p.terminate()
    p.wait()
    exit(0)
  signal.signal(signal.SIGTERM,sigterm)

  q.put(('ready',i))

  p.stdin.write(G)
  p.stdin.flush()
  data = ''
  for line in p.stdout:
    data += line
  if data == '':
    q.put(('maxiterations',data))
  else:
    q.put(('success',data))
  while True:
    time.sleep(1)

# multiprocessing forks python reasonably quickly
# and then the python processes start the binary in parallel
# whereas serializing subprocess.Popen would have relatively high latency

q = multiprocessing.Queue()
p = [multiprocessing.Process(target=lowweight,args=(i,)) for i in range(threads)]
for i in range(threads): p[i].start()

# try to avoid deadlocks for very fast runs

jobsrunning = threads
ready = [False]*threads
success = None
while jobsrunning > 0:
  cmd,data = q.get()
  if cmd == 'ready':
    ready[data] = True
  if cmd == 'maxiterations':
    jobsrunning -= 1
  if cmd == 'success':
    success = data
    jobsrunning -= 1
  if all(ready) and success != None:
    break

for i in range(threads): p[i].terminate()
for i in range(threads): p[i].join()
if success != None:
  sys.stdout.write(success)
